#pragma once

namespace nifty{
namespace graph{
namespace opt{
namespace minstcut{

    template<class OBJECTIVE>
    struct MinstcutObjectiveName;

    // template<class GRAPH>
    // struct GraphName{
    //     static std::string name(){
    //
    //     }
    // };

} // namespace minstcut
} // namespace opt
}
}

