#pragma once

namespace nifty{
namespace graph{
namespace opt{
namespace multicut{

    template<class OBJECTIVE>
    struct MulticutObjectiveName;
    
    // template<class GRAPH>
    // struct GraphName{
    //     static std::string name(){
    // 
    //     }
    // };

} // namespace multicut 
} // namespace opt   
}
}

