#pragma once

namespace nifty{
namespace graph{

    template<class GRAPH>
    struct GraphName;

    // template<class GRAPH>
    // struct GraphName{
    //     static std::string name(){
    // 
    //     }
    //     static std::string moduelName(){
    // 
    //     }
    // };
    
}
}

