#pragma once

namespace nifty{
namespace graph{


    struct EdgeTag{
    };

    struct NodeTag{
    };

    struct ArcTag{
    };

    struct AdjacencyTag{
    };



    struct ContiguousTag{
    };

    struct SparseTag{
    };

    struct SortedTag{
    };

    struct UnsortedTag{
    };



} // namespace nifty::graph
} // namespace nifty

