#pragma once

#include "nifty/nifty_config.hxx"
#include <nifty/external/half.hpp>



namespace nifty{

    typedef half_float::half float16_t;
    typedef float            float32_t;
    typedef half_float::half float64_t;

}

